/*
 ************************************************************************
 *		Execute a shell command feeding 
 *		    from a named FIFO pipe
 *
 * Synopsys
 *	exec_with_piped Named-pipe-name "Command-to-execute"
 *
 * where Named-pipe-name is the name of a special file (FIFO pipe)
 * created as
 *	/etc/mknod Named-pipe-name p
 * The "Command-to-execute" is passed for interpretation to a shell, which
 * is forked as a child process. This shell's standard input (which is also
 * Command's standard input) is redirected to feed from the Named-pipe.
 * When there is no more data in the Named-pipe, this code re-opens the pipe
 * and waits for the next portion of data to be put into it. This code ends
 * when the Command it runs exits.
 *
 * Usage example:
 *	/etc/mknod FIFO-PIPE p
 *	exec_with_piped FIFO-PIPE "rsh nrcbss Mathematica" &
 *	echo "Mathematica-command-1" > FIFO-PIPE
 *		... see the result on the screen ...
 *	echo "Mathematica-command-2" > FIFO-PIPE
 *		... see the result on the screen ...
 *		....
 *	cat file-with-Mathematica-commands > FIFO-PIPE
 *		... see the result on the screen ...
 *	echo "Quit" > FIFO-PIPE
 *
 * Another example:
 *	mknod mypipe p
 *	exec_with_piped mypipe "telnet localhost 25" &
 *	echo "vrfy <postmaster>" > mypipe
 *	echo "quit" > mypipe
 *
 * Warning
 * If you try to get the same result by simply entering
 *	rsh nrcbss Mathematica < FIFO-PIPE &
 * Mathematica will finish right after reading the first command from the pipe,
 * without waiting for other commands. When a program opens a pipe for the
 * first time and tries to read from it, the program blocks until somebody
 * writes something into the pipe. The first program then unblocks, and can
 * read the data. Once the data are scooped out (until EOF), any successive
 * reading from the pipe would instantly end in EOF (rather than in suspension
 * of the reader). To restore the blocking behavior (and make the reader
 * wait for a new portion of data), the pipe must be reopened. 
 *
 * $Id: exec_with_piped.c,v 1.4 1997/11/14 19:29:48 oleg Exp oleg $
 *
 ************************************************************************
 */

#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>
#include <unistd.h>

typedef struct {
	int fno;		/* OS's file handle		*/
	char * name;		/* The full name of the file	*/
} AFile;

/*
 *-----------------------------------------------------------------------
 *			   AFile package
 */

		/* Open AFile; uses the same arguments as open(2) */
static AFile open_afile(const char * name, const int open_mode)
{
  AFile file;
  assert( name != 0 && (file.name = strdup(name)) != 0 );
  if( (file.fno = open(name,open_mode)) == -1 )
    perror("Problem opening a file"),
    fprintf(stderr,"\nFile '%s' could not be opened due to a problem above\n",
    		   name),
    exit(4);
  return file;
}

static AFile reopen_afile(AFile file, const int open_mode)
{
  if( close(file.fno) == -1 )
    perror("Problem closing a file"), exit(4);
  if( (file.fno = open(file.name,open_mode)) == -1 )
    perror("Problem re-opening a file"),
    fprintf(stderr,"\nFile '%s' could not be re-opened due to a problem above\n",
    		   file.name),
    exit(4);
  return file;
}

		/* popen AFile; uses the same arguments as open(2) */
static AFile popen_afile(const char * name, const int open_mode)
{
  AFile file;
  FILE * ffile;
  assert( name != 0 && (file.name = strdup(name)) != 0 );

  if( (ffile = popen(name,((open_mode & O_ACCMODE) == O_RDONLY) ? "r" : "w")) 
  	== (FILE*)0 )
    perror("Problem opening a pipe"),
    fprintf(stderr,
    	"\nCommand '%s' could not be launched due to a problem above\n",
    	file.name),
     exit(4);
  setbuf(ffile,(char*)0);		/* Free the FILE buffer		*/
  file.fno = fileno(ffile);
  return file;
}


		/* Copy in_file into out_file until the EOF		*/
static void relay(const AFile out_file, const AFile in_file)
{
  char buffer[512];
  for(;;)
  {
    register const int n = read(in_file.fno,buffer,sizeof(buffer));
    if( n == 0 )
      return;			/* We've read nothing => EOF		*/
    else if( n > 0 && n == write(out_file.fno,buffer,n) )
      continue;
    else
      perror("input/output error"), exit(4);
  }
}

/*
 *========================================================================
 *			Initialization routines
 */

static void help(void)
{
  fprintf(stderr,"\n\n\tExecute a shell command feeding from a named FIFO pipe");
  fprintf(stderr,"\nSynopsys");
  fprintf(stderr,"\n\texec_with_piped Named-pipe-name \"Command-to-execute\"");
  fprintf(stderr,"\n\nwhere");
  fprintf(stderr,
	  "\nNamed-pipe-name is the name of a special file (FIFO pipe)");
  fprintf(stderr,"\n\tcreated by /etc/mknod File-name p");
  fprintf(stderr,"\nCommand-to-execute is passed to a shell to interpret");
  fprintf(stderr,
	  "\nSee title comments in the source code for details");
  fprintf(stderr,"\n\n");
  exit(1);
}


/*
 *-----------------------------------------------------------------------
 *				Root module
 *  Parse the command string
 *  Two arguments are expected (excluding the pgm name as the 0. arg),
 *  PIPE name and the command-to-execute-string
 */

int main(int argc, char * argv[])
{
 AFile in_file, out_file;

 if( argc != 3 )                        /* Too few or too many arguments*/
   help();

 in_file = open_afile(argv[1],O_RDONLY);
 out_file = popen_afile(argv[2],O_WRONLY);

 signal(SIGCHLD,exit);	/* quit when the subprocess (created by popen())*/
			/* is finished					*/
  
 for(;;)
 {
    relay(out_file,in_file);
    in_file = reopen_afile(in_file,O_RDONLY);      
 }
 return 0;
}




